package phylonet.coalescent;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Stack;
import java.util.TreeSet;
import java.util.Map.Entry;

import phylonet.coalescent.Utils.ClusterComparator;
import phylonet.lca.SchieberVishkinLCA;
import phylonet.tree.model.MutableTree;
import phylonet.tree.model.TMutableNode;
import phylonet.tree.model.TNode;
import phylonet.tree.model.Tree;
import phylonet.tree.model.sti.STINode;
import phylonet.tree.model.sti.STITree;
import phylonet.tree.model.sti.STITreeCluster;
import phylonet.tree.util.PostTraversal;
import phylonet.tree.util.Trees;
import phylonet.util.BitSet;

public class RecursiveGreedyInference {

	
	//protected List<Tree> trees;
	private HashMap<BitSet, Integer> counts;
	private TaxonIdentifier leaves;
	private TreeSet<Entry<BitSet,Integer>> countSorted = null;
	private int n;
	private int k;
	private int level = 0;
	private String outgroup;
	
	public RecursiveGreedyInference(List<Tree> trees) {
		this.checkInputTrees(trees);
		this.setBitSetCounts(trees);
	}

	private RecursiveGreedyInference(List<String> randomSample, int k) {
		this.n = randomSample.size();
		this.k = k;
		this.counts = new HashMap<BitSet, Integer>();
		this.leaves = new TaxonIdentifier();
		for (String sample : randomSample) {
			this.leaves.taxonId(sample);
		}
		this.leaves.lock();
	}

	private void checkInputTrees(List<Tree> trees) {
		leaves = new TaxonIdentifier();
		k = trees.size();
		String[] firstTreeLeaves = trees.iterator().next().getLeaves();
		for (String leaf : firstTreeLeaves) {
			leaves.taxonId(leaf);
		}
		leaves.lock();
		n = leaves.taxonCount();
		this.outgroup = firstTreeLeaves[firstTreeLeaves.length - 1];
		
		int t = 0;
		for (Tree tree: trees) {
			if (tree.getLeafCount() != n) {
				throw new RuntimeException("Tree "+t+" does not include all taxa from previous trees");
			}
			for (String leaf: tree.getLeaves()) {
				try{
					leaves.taxonId(leaf);
				} catch (RuntimeException e) {
					throw new RuntimeException("Leaf " + leaf + " in tree" + t +
							"not present in previous trees: "+leaf);
				}
			}
			t++;
		}
	}

	
	public List<Solution> inferSpeciesTree() {
		long startTime = System.currentTimeMillis();

		List<Solution> solutions = new ArrayList<Solution>();
		Solution sol = new Solution();
		sol._st = this.recuresiveGreedy(1/3d, true);
		System.err.println("resolution: " + (sol._st.getNodeCount() - n - 2) + " out of " + (n -3));
		solutions.add(sol);

		/*if (GlobalMaps.taxonNameMap == null && rooted && extraTrees == null && false) {
			restoreCollapse(solutions, cd);
		}*/
		
		/*System.out.println("Inference took " + (System.currentTimeMillis()  - startTime)/1000
				+ " seconds");*/

		return  solutions;
	}
	
	private List<BitSet> bitSetsForGene(Tree fulltree) {
	
		ArrayList<BitSet> ret = new ArrayList<BitSet>();
	
		Stack<BitSet> stack = new Stack<BitSet>();
		for (TNode rgtn : fulltree.postTraverse()) {
	
			if (rgtn.isRoot() && rgtn.getChildCount() == 2) {
				continue;
			}
			BitSet bs = new BitSet(n);
			if (rgtn.isLeaf()) {
				int i =  leaves.taxonId(rgtn.getName());               
				bs.set(i); 
			} else {
				int childCount = rgtn.getChildCount();
				for (int i = 0; i < childCount; i++) {
					BitSet pop = stack.pop();
					bs.or(pop);
				}
			}
			stack.push(bs);
			int bsc = bs.cardinality();
			if (bsc < 2 || bsc >= n - 1) {
				continue;
			}       
			ret.add(bs);
		}
		return ret;
	}


	private void addToCounts(List<BitSet> bsList) {
		for (BitSet bs : bsList) {
			if (counts.containsKey(bs)) {
				counts.put(bs, counts.get(bs) + 1);
				continue;
			}
			BitSet bs2 = (BitSet)bs.clone();
			bs2.flip(0,n);
			if (counts.containsKey(bs2)) {
				counts.put(bs2, counts.get(bs2) + 1);
				continue;
			}
			counts.put(bs, 1);
		}
	}

	private void setBitSetCounts(List<Tree> trees) {		
		counts = new HashMap<BitSet, Integer>();		
		for (Tree gt : trees) {		
			gt.rerootTreeAtEdge(this.outgroup);
			addToCounts(this.bitSetsForGene(gt));
		}
	}

	private void resort(boolean randomize) {
		countSorted = new TreeSet<Entry<BitSet,Integer>>(new BSComparator(randomize, n));
        countSorted.addAll(counts.entrySet());
	}

	public Tree buildTreeFromBitSets(Iterable<BitSet> bitSets) {
		ArrayList<STITreeCluster> clusters = new ArrayList<STITreeCluster>();
		for (BitSet bs: bitSets) {
			STITreeCluster cluster = new STITreeCluster(this.leaves);
			cluster.setCluster(bs);
			//System.err.println(cluster.getClusterSize());
			clusters.add(cluster);
		}
		return this.buildTreeFromClusters(clusters);
	}
	
	public Tree buildTreeFromClusters(Iterable<STITreeCluster> clusters) {
        if (clusters == null) {
          throw new RuntimeException("Empty list of clusters. The function returns a null tree.");
        }
    
        MutableTree tree = new STITree<Double>();
    
        //String[] taxa = ((STITreeCluster)clusters.get(0)).getTaxa();
        for (int i = 0; i < this.leaves.taxonCount(); i++) {
          tree.getRoot().createChild(this.leaves.getTaxonName(i));
        }
    
        for (STITreeCluster tc : clusters) {
          if ((tc.getClusterSize() <= 1) || (tc.getClusterSize() == n))
          {
            continue;
          }
    
          Set<TNode> clusterLeaves = new HashSet<TNode>();
          TNode node;
          for (String l : tc.getClusterLeaves()) {
            node = tree.getNode(l);
            clusterLeaves.add(node);
          }
    
          SchieberVishkinLCA lcaFinder = new SchieberVishkinLCA(tree);
          TNode lca = lcaFinder.getLCA(clusterLeaves);
    
          LinkedList<TNode> movedChildren = new LinkedList<TNode>();
          int nodes = clusterLeaves.size();
          for (TNode child : lca.getChildren()) {
            BitSet childCluster = new BitSet(n);
            for (TNode cl : child.getLeaves()) {
              int i = leaves.taxonId(cl.getName());
              childCluster.set(i);
            }
            
    
            BitSet temp = (BitSet)childCluster.clone();
            temp.and(tc.getBitSet());
            if (temp.equals(childCluster)) {
              movedChildren.add(child);
              nodes -= temp.cardinality();
            }
    
          }
          
          if (movedChildren.size() == 0 || nodes != 0) {
              continue;
          }
    
          STINode newChild = ((STINode)lca).createChild();
    
          while (!movedChildren.isEmpty()) {
            newChild.adoptChild((TMutableNode)movedChildren.get(0));
            movedChildren.remove(0);
          }
        }
    
        ((STITree<Double>)tree).setRooted(false);
        return (Tree)tree;
      }
	
	public final List<Tree> greedyConsensus(double [] thresholds, boolean randomize, int repeate) {
		List<Tree> outtrees = new ArrayList<Tree>();
		for (int i = 0; i < repeate; i ++) {
			outtrees.addAll(this.greedyConsensus(thresholds, true, randomize));
		}
		return outtrees;
	}
	
	public final List<Tree> greedyConsensus(double [] thresholds, boolean resort, boolean randomize) {

		if (resort || this.countSorted == null) {
			this.resort(randomize);
		}

		List<Tree> outtrees = new ArrayList<Tree>();
		int ti = thresholds.length - 1;
        double threshold = thresholds[ti];
		List<BitSet> bitSets = new ArrayList<BitSet>();   
		for (Entry<BitSet, Integer> entry : countSorted) {
			if (threshold > (entry.getValue()+.0d)/k) {	
				outtrees.add(this.buildTreeFromBitSets(bitSets));
				ti--;
        		if (ti < 0) {
        			break;
        		}
        		threshold = thresholds[ti];
			}
			bitSets.add(entry.getKey());
		}
        while (ti >= 0) {
    		outtrees.add(this.buildTreeFromBitSets(bitSets));
    		ti--;
        }
        return outtrees;
	}

	
	public Tree greedyConsensus(double threshold, boolean resort,
			boolean randomize) {
		return this.greedyConsensus(new double[]{threshold}, resort, randomize).get(0);
	}

	private RecursiveGreedyInference induce(List<String> randomSample) {
		
		int sampleSize = randomSample.size();
		
		RecursiveGreedyInference induced = new RecursiveGreedyInference(randomSample,k);
		induced.outgroup = randomSample.get(randomSample.size() - 1);
		
		Integer outgroupId = this.leaves.taxonId(outgroup);
		
		List<BitSet> inducedBsList = new ArrayList<BitSet>(this.countSorted.size());
		for (Entry<BitSet, Integer> entry: this.counts.entrySet()) {
			
			BitSet bs = entry.getKey();
			boolean bsHasOut = bs.get(outgroupId);
			
			BitSet inducedbs = new BitSet(sampleSize);
			int s = 0;
			for (int i = 0; i < sampleSize; i++) {
				boolean isSet = bs.get(this.leaves.taxonId(randomSample.get(i)));
				if ((bsHasOut && !isSet) || (!bsHasOut && isSet)) {
					inducedbs.set(i);
					s++;
				}
			}
			if (s > 1 && s < sampleSize - 1)
				inducedBsList.add(inducedbs);
		}
		induced.addToCounts(inducedBsList);
		induced.level = this.level + 1;
		return induced;
	}
	
	private long  randomSampleAroundPolytomy(ArrayList<TNode> children, 
			List<String> randomSample) {
		Set<String> allLeaves = new HashSet<String>();
		long possibilities = 1;
		for (TNode child : children) {
			int sample = GlobalMaps.random.nextInt(child.getLeafCount());
			long i = 0;
			for (TNode leaf: child.getLeaves()) {
				if (i == sample) {
					randomSample.add(leaf.getName());
				}
				i++;
				allLeaves.add(leaf.getName());
			}
			possibilities *= i;
		}
		int sample = GlobalMaps.random.nextInt(n - allLeaves.size());
		for (int i = 0; sample >= 0; i++) {
			if (!allLeaves.contains(this.leaves.getTaxonName(i))) {
				if (sample == 0) {
					randomSample.add(this.leaves.getTaxonName(i));
				}
				sample --;
			}
		}
		possibilities *= n - allLeaves.size();
		return possibilities;
	}
	
	private void resolveNodeFromInduced(TNode node, ArrayList<TNode> children,
			RecursiveGreedyInference inducedGreedy, Tree resolution) {
		Stack<TMutableNode> stack = new Stack<TMutableNode>();
		for (TNode resNode: resolution.postTraverse()) {
			if (resNode.isLeaf()) {
				Integer childId = inducedGreedy.leaves.taxonId(resNode.getName());
				if (childId < children.size()) {
					TNode child = children.get(childId);
					stack.push((TMutableNode) child);
				} else {
					// This is the "root"; i.e. everything else. 
				}
			} else {
				if (resNode.isRoot() || resNode.getParent().isRoot()) {
					break;
				}
				TMutableNode newNode = ((TMutableNode)node).createChild();
				for (int i = 0; i < resNode.getChildCount(); i++) {
					newNode.adoptChild(stack.pop());
				}
				stack.push(newNode);
			}
			
		}
	}

	private void print(String string) {
		for (int l = -1; l < this.level; l++)
			System.err.print("-");
		System.err.println(string);
	}
	
	public Tree recuresiveGreedy(double threshold, boolean randomize) {
		Tree greedy = this.greedyConsensus(threshold, true, randomize);
		greedy.rerootTreeAtEdge(leaves.getTaxonName(leaves.taxonCount()-1));
		
		if (greedy.getNodeCount() == n + 2) {
			print(" X maximum freq is only " + (this.countSorted.first().getValue()+0.d)/k);
			return greedy;
		}
		for (TNode node: greedy.postTraverse()) {
			if (node.getChildCount() > 2 && ! node.isRoot()) {
				int tries = 0;
				print(" + polytomy of size " + node.getChildCount()+ ": "+node);
				
				while (tries < 100){

					ArrayList<TNode> children = Utils.getChildrenAsList(node);
					List<String> sample = new ArrayList<String>();
					long ways = this.randomSampleAroundPolytomy(children, sample);
					if (tries > ways || ways == 1) {
						this.print(" X no (more) sampling possible for " + node 
								+ " tried " + tries + " of " + ways);
						break;
					}
					RecursiveGreedyInference inducedGreedy = this.induce(sample);
					Tree resolution = inducedGreedy.recuresiveGreedy(threshold, randomize);
					if (resolution.getNodeCount() == inducedGreedy.n + 2) {
						this.print(" . no extra resolution");
					} else {
						this.print(" . further resolved: " + node+" -> "+resolution);
						resolveNodeFromInduced(node, children, inducedGreedy, resolution);
					}
					if (node.getChildCount() == 2) {
						break;
					}
					tries ++;
					print(" (tried "+tries+ " out of possible " +ways + " with "+sample+")");
				}
			}
		}
		return greedy;
	}

	public static class BSComparator implements Comparator<Entry<BitSet,Integer>> {

		//private boolean random;
		List<Integer> inds;
		public BSComparator (boolean randomize, int size) {
			inds = new ArrayList<Integer>(); 
			for (int i = 0; i < size; i++) {
				inds.add(i);
			}
			if (randomize) {
				Collections.shuffle(inds, GlobalMaps.random);
			}
		}
		@Override
		public int compare(Entry<BitSet, Integer> o1,
				Entry<BitSet, Integer> o2) {
			return compare(o1.getKey(),o1.getValue(), o2.getKey(), o2.getValue());
		}
		private int compare(BitSet k1, Integer v1, BitSet k2, Integer v2) {
			int a = v2.compareTo(v1);
			if (a != 0) {
				return a;
			}
			if  (k1.equals(k2)) {
				return 0;
			}
			for (int ind : this.inds) {
				boolean j = k1.get(ind);
				boolean jj = k2.get(ind);
				if (j != jj) {
					return (j) ? 1 : -1;
				} 
			}
			throw new RuntimeException("hmm! this should never be reached");
		}
	}
}
